// SPDX-License-Identifier: Unlicensed
pragma solidity >=0.8.0 <0.9.0;

import "./IAdditionalSecurityModule.sol";

// Delay ASM
contract DelayASM is IAdditionalSecurityModule {
    mapping(bytes32 => bool) public delayedTransfers;
    address public gatewayContract;
    uint256 public delayPeriod;
    address public owner;
    string public appRouterBridgeAddress;

    constructor(address gatewayAddress, uint256 _delayPeriod, string memory _routerBridgeAddress) {
        gatewayContract = gatewayAddress;
        owner = msg.sender;
        delayPeriod = _delayPeriod;
        appRouterBridgeAddress = _routerBridgeAddress;
    }

    function setDelayPeriod(uint256 _delayPeriod) external {
        require(msg.sender == owner, "Caller is not owner");
        delayPeriod = _delayPeriod;
    }

    function rejectRequest(bytes32 id) external {
        require(msg.sender == owner);
        delayedTransfers[id] = true;
    }

    function isRequestRejected(bytes32 id) internal view returns (bool) {
        return delayedTransfers[id];
    }

    function _validateDelay(uint256 requestTimestamp) internal view returns (bool) {
        if (block.timestamp > requestTimestamp + delayPeriod) {
            return true;
        } else {
            revert("Transaction needs to be delayed");
        }
    }

    function verifyCrossChainRequest(
        uint256 eventNonce,
        uint256 requestTimestamp,
        string calldata srcContractAddress,
        string calldata srcChainId,
        bytes calldata payload,
        address handler
    ) external view returns (bool) {
        require(msg.sender == gatewayContract, "Caller is not gateway");
        bytes32 id = keccak256(
            abi.encode(eventNonce, requestTimestamp, srcContractAddress, srcChainId, payload, handler)
        );
        bool isRejected = isRequestRejected(id);
        if (isRejected) {
            return !isRejected;
        }
        _validateDelay(requestTimestamp);
        return !isRejected;
    }
}
